# -*- coding: utf-8 -*-
import os
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup

cid = vim.channel_id


def host_setup():
    cleanup()
    # Spawn the python host
    vim.command(
        'let pyhost_id = ' +
        'api_spawn("python", ["-c", "import neovim; neovim.start_host()"])')
    ok(vim.eval('g:pyhost_id'))
    # Use send_call to wait for the host setup(api_spawn will return a channel
    # id but only after a while the channel will register handlers for python_*
    # methods)
    ok(vim.eval('send_call(g:pyhost_id, "python_eval", "10")') == 10)
    # Verify the feature
    ok(vim.eval('has("python")'))
    # Import the vim module
    vim.command('python import vim')
    # Ensure the python host was updated accordingly
    ok(vim.eval('pyeval("vim.channel_id") == g:pyhost_id'))


def host_teardown():
    ok(vim.eval('api_close(g:pyhost_id)'))
    # After the channel is closed, the feature should not be available
    ok(not vim.eval('has("python")'))


@with_setup(setup=host_setup, teardown=host_teardown)
def test_python_command():
    vim.command('python vim.command("let set_by_python = [100,0]")')
    eq(vim.vars['set_by_python'], [100, 0])


@with_setup(setup=host_setup, teardown=host_teardown)
def test_python_nested_commands():
    nested = """python vim.command('python vim.command("python vim.command(\\'let set_by_nested_python = 555\\')")')"""
    vim.command(nested)
    eq(vim.vars['set_by_nested_python'], 555)


@with_setup(setup=host_setup, teardown=host_teardown)
def test_pyfile():
    fname = 'pyfile.py'
    text = 'vim.command("let set_by_pyfile = 123")'
    with open(fname, 'w') as f: f.write(text)
    vim.command('pyfile pyfile.py')
    eq(vim.vars['set_by_pyfile'], 123)
    os.unlink(fname)


@with_setup(setup=host_setup, teardown=host_teardown)
def test_pydo():
    # insert some text
    vim.command('normal iabc\ndef\nghi')
    eq(vim.current.buffer[:], ['abc', 'def', 'ghi'])
    # go to top and select and replace the first two lines
    vim.command('normal ggvj:pydo return linenr\n')
    eq(vim.current.buffer[:], ['1', '2', 'ghi'])


@with_setup(setup=host_setup, teardown=host_teardown)
def test_pyeval():
    vim.command('let python_expr = pyeval("[1, 2, 3]")')
    eq(vim.vars['python_expr'], [1, 2, 3])



from collections import deque
import logging

logger = logging.getLogger(__name__)
debug, info, warn = (logger.debug, logger.info, logger.warn,)

class RPCStream(object):
    def __init__(self, stream):
        self.stream = stream
        self.pending_requests = {}
        self.next_request_id = 1
        self.interrupted = False
        self.stopped = False
        self.running = False
        self.posted_notifications = deque()


    def configure(self, vim):
        self.stream.configure(vim)


    def post(self, name, args):
        self.posted_notifications.append((name, args,))
        self.stream.interrupt()


    def send(self, method, args, response_cb):
        request_id = self.next_request_id
        # Update request id
        self.next_request_id = request_id + 1
        # Send the request
        self.stream.send([0, request_id, method, args])
        # set the callback
        self.pending_requests[request_id] = response_cb


    def loop_start(self, request_cb, notification_cb, error_cb):
        def msg_cb(msg):
            msg_type = msg[0]
            if msg_type == 0:
                # request
                #   - msg[1]: id
                #   - msg[2]: method name
                #   - msg[3]: arguments
                pass # replaces next logging statement
                #debug('received request: %s, %s', msg[2], msg[3])
                request_cb(msg[2], msg[3], reply_fn(self.stream, msg[1]))
            elif msg_type == 1:
                # response to a previous request:
                #   - msg[1]: the id
                #   - msg[2]: error(if any)
                #   - msg[3]: result(if not errored)
                pass # replaces next logging statement
                #debug('received response: %s, %s', msg[2], msg[3])
                self.pending_requests.pop(msg[1])(msg[2], msg[3])
            elif msg_type == 2:
                # notification/event
                #   - msg[1]: event name
                #   - msg[2]: arguments
                pass # replaces next logging statement
                #debug('received notification: %s, %s', msg[1], msg[2])
                notification_cb(msg[1], msg[2])
            else:
                error = 'Received invalid message %s' % msg
                pass # replaces next logging statement
                #warn(error)
                raise Exception(error)

        self._run(msg_cb, notification_cb, error_cb)
        pass # replaces next logging statement
        #debug('exiting rpc stream loop')



    def loop_stop(self):
        self.stopped = True
        if self.running:
            self.stream.loop_stop()


    def _run(self, msg_cb, notification_cb, error_cb):
        self.stopped = False
        while not self.stopped:
            if self.posted_notifications:
                notification_cb(*self.posted_notifications.popleft())
                continue
            self.running = True
            self.stream.loop_start(msg_cb, error_cb)
            self.running = False


def reply_fn(stream, request_id):
    def reply(value, error=False):
        if error:
            resp = [1, request_id, value, None]
        else:
            resp = [1, request_id, None, value]
        stream.send(resp)

    return reply



from imp import new_module, find_module, load_module
import sys, logging, os.path
from traceback import format_exc

logger = logging.getLogger(__name__)
debug, warn = (logger.debug, logger.warn,)

class ScriptHost(object):
    """
    Plugin that provides the 'python' feature, emulating an environment for
    python code similar to the one provided by vim-python bindings.
    """
    def __init__(self, vim):
        self.provides = ['python']
        self.vim = vim
        # context where all code will run
        self.module = new_module('__main__')
        vim.script_context = self.module
        # it seems some plugins assume 'sys' is already imported, so do it now
        exec 'import sys' in self.module.__dict__

    def python_execute(self, script):
        exec script in self.module.__dict__

    def python_execute_file(self, file_path):
        execfile(file_path, self.module.__dict__)

    def python_do_range(self, start, stop, code):
        vim = self.vim
        start -= 1
        stop -= 1
        fname = '_vim_pydo'
        # define the function
        function_def = 'def %s(line, linenr):\n %s' % (fname, code,)
        exec function_def in self.module.__dict__
        # get the function
        function = self.module.__dict__[fname]
        while start <= stop:
            # Process batches of 5000 to avoid the overhead of making multiple
            # API calls for every line. Assuming an average line length of 100
            # bytes, approximately 488 kilobytes will be transferred per batch,
            # which can be done very quickly in a single API call.
            sstart = start
            sstop = min(start + 5000, stop)
            lines = vim.current.buffer.get_slice(sstart, sstop, True, True)
            for i, line in enumerate(lines):
                linenr = i + sstart + 1
                result = str(function(line, linenr))
                if result:
                    lines[i] = result
            start = sstop + 1
            vim.current.buffer.set_slice(sstart, sstop, True, True, lines)
        # delete the function
        del self.module.__dict__[fname]

    def python_eval(self, expr):
        return eval(expr, self.module.__dict__)




# -*- coding: utf-8 -*-
import os
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup

cid = vim.channel_id


@with_setup(setup=cleanup)
def test_call_and_reply():
    errors = []

    def notification_cb(name, args):
        eq(name, 'setup')
        cmd = 'let g:result = send_call(%d, "client-call", 1, 2, 3)' % cid
        vim.command(cmd)
        eq(vim.vars['result'], [4, 5, 6])
        vim.loop_stop()

    def request_cb(name, args):
        eq(name, 'client-call')
        eq(args, [1, 2, 3])
        return [4, 5, 6]

    def error_cb(err):
        errors.append(err)
        vim.loop_stop()

    vim.post('setup')
    vim.loop_start(request_cb, notification_cb, error_cb)
    if errors:
        raise errors[0]


@with_setup(setup=cleanup)
def test_call_api_before_reply():
    errors = []

    def notification_cb(name, args):
        eq(name, 'setup2')
        cmd = 'let g:result = send_call(%d, "client-call2", 1, 2, 3)' % cid
        vim.command(cmd)
        eq(vim.vars['result'], [7, 8, 9])
        vim.loop_stop()

    def request_cb(name, args):
        eq(name, 'client-call2')
        eq(args, [1, 2, 3])
        vim.command('let g:result2 = [7, 8, 9]')
        return vim.vars['result2']

    def error_cb(err):
        errors.append(err)
        vim.loop_stop()

    vim.post('setup2')
    vim.loop_start(request_cb, notification_cb, error_cb)
    if errors:
        raise errors[0]


@with_setup(setup=cleanup)
def test_recursion():
    errors = []

    def notification_cb(name, args):
        eq(name, 'setup3')
        vim.vars['result1'] = 0
        vim.vars['result2'] = 0
        vim.vars['result3'] = 0
        vim.vars['result4'] = 0
        cmd = 'let g:result1 = send_call(%d, "call", %d)' % (cid, 2,)
        vim.command(cmd)
        eq(vim.vars['result1'], 4)
        eq(vim.vars['result2'], 8)
        eq(vim.vars['result3'], 16)
        eq(vim.vars['result4'], 32)
        vim.loop_stop()

    def request_cb(name, args):
        n = args[0]
        n *= 2
        if n <= 16:
            if n == 4:
                cmd = 'let g:result2 = send_call(%d, "call", %d)' % (cid, n,)
            elif n == 8:
                cmd = 'let g:result3 = send_call(%d, "call", %d)' % (cid, n,)
            elif n == 16:
                cmd = 'let g:result4 = send_call(%d, "call", %d)' % (cid, n,)
            vim.command(cmd)
        return n

    def error_cb(err):
        errors.append(err)
        vim.loop_stop()

    vim.post('setup3')
    vim.loop_start(request_cb, notification_cb, error_cb)
    if errors:
        raise errors[0]




# -*- coding: utf-8 -*-
import os, tempfile
from nose.tools import with_setup, eq_ as eq, ok_ as ok
from common import vim, cleanup


@with_setup(setup=cleanup)
def test_command():
    fname = tempfile.mkstemp()[1]
    vim.command('new')
    vim.command('edit %s' % fname)
    vim.command('normal itesting\npython\napi')
    vim.command('w')
    ok(os.path.isfile(fname))
    eq(file(fname).read(), 'testing\npython\napi\n')
    os.unlink(fname)


@with_setup(setup=cleanup)
def test_eval():
    vim.command('let g:v1 = "a"')
    vim.command('let g:v2 = [1, 2, {"v3": 3}]')
    eq(vim.eval('g:'), {'v1': 'a', 'v2': [1, 2, {'v3': 3}]})


@with_setup(setup=cleanup)
def test_strwidth():
    eq(vim.strwidth('abc'), 3)
    # 6 + (neovim)
    # 19 * 2 (each japanese character occupies two cells)
    eq(vim.strwidth('neovimのデザインかなりまともなのになってる。'), 44)


@with_setup(setup=cleanup)
def test_list_runtime_paths():
    # Is this the default runtime path list?
    homedir = os.environ['HOME'] + '/.nvim'
    dflt_rtp = [
        homedir,
        '/usr/local/share/nvim/vimfiles',
        '/usr/local/share/nvim',
        '/usr/local/share/nvim/vimfiles/after'
    ]
    # If the runtime is installed the default path
    # is nvim/runtime
    dflt_rtp2 = list(dflt_rtp)
    dflt_rtp2[2] += '/runtime'

    rtp = vim.list_runtime_paths()
    ok(rtp == dflt_rtp or rtp == dflt_rtp2)



@with_setup(setup=cleanup)
def test_chdir():
    pwd = vim.eval('getcwd()')
    vim.chdir('/')
    eq(vim.eval('getcwd()'), '/')
    vim.chdir(pwd)
    eq(vim.eval('getcwd()'), pwd)


@with_setup(setup=cleanup)
def test_current_line():
    eq(vim.current.line, '')
    vim.current.line = 'abc'
    eq(vim.current.line, 'abc')


@with_setup(setup=cleanup)
def test_vars():
    vim.vars['python'] = [1, 2, {'3': 1}]
    eq(vim.vars['python'], [1, 2, {'3': 1}])
    eq(vim.eval('g:python'), [1, 2, {'3': 1}])


@with_setup(setup=cleanup)
def test_options():
    eq(vim.options['compatible'], True)
    vim.options['compatible'] = False
    eq(vim.options['compatible'], False)


@with_setup(setup=cleanup)
def test_buffers():
    eq(len(vim.buffers), 1)
    eq(vim.buffers[0], vim.current.buffer)
    vim.command('new')
    eq(len(vim.buffers), 2)
    eq(vim.buffers[1], vim.current.buffer)
    vim.current.buffer = vim.buffers[0]
    eq(vim.buffers[0], vim.current.buffer)


@with_setup(setup=cleanup)
def test_windows():
    eq(len(vim.windows), 1)
    eq(vim.windows[0], vim.current.window)
    vim.command('vsplit')
    vim.command('split')
    eq(len(vim.windows), 3)
    eq(vim.windows[0], vim.current.window)
    vim.current.window = vim.windows[1]
    eq(vim.windows[1], vim.current.window)


@with_setup(setup=cleanup)
def test_tabpages():
    eq(len(vim.tabpages), 1)
    eq(vim.tabpages[0], vim.current.tabpage)
    vim.command('tabnew')
    eq(len(vim.tabpages), 2)
    eq(len(vim.windows), 2)
    eq(vim.windows[1], vim.current.window)
    eq(vim.tabpages[1], vim.current.tabpage)
    vim.current.window = vim.windows[0]
    # Switching window also switches tabpages if necessary(this probably
    # isn't the current behavior, but compatibility will be handled in the
    # python client with an optional parameter)
    eq(vim.tabpages[0], vim.current.tabpage)
    eq(vim.windows[0], vim.current.window)
    vim.current.tabpage = vim.tabpages[1]
    eq(vim.tabpages[1], vim.current.tabpage)
    eq(vim.windows[1], vim.current.window)



import msgpack, logging

logger = logging.getLogger(__name__)
debug, warn = (logger.debug, logger.warn,)

class MsgpackStream(object):
    def __init__(self, stream):
        def default(obj):
            return obj._handle

        def ext_hook(code, data):
            klass = self.types[code]
            rv = klass(self.vim, msgpack.ExtType(code, data))
            klass.initialize(rv)
            return rv

        self.types = None
        self.vim = None
        self.packer = msgpack.Packer(use_bin_type=True, default=default)
        self.unpacker = msgpack.Unpacker(ext_hook=ext_hook)
        self.stream = stream


    def configure(self, vim):
        self.vim = vim
        self.types = {}
        for name, info in vim.api_metadata['types'].items():
            self.types[info['id']] = getattr(vim, name)


    def interrupt(self):
        self.stream.interrupt()


    def send(self, msg):
        self.stream.send(self.packer.pack(msg))


    def loop_start(self, msg_cb, error_cb):
        def data_cb(data):
            self.unpacker.feed(data)

            while True:
                try:
                    pass # replaces next logging statement
                    #debug('waiting for message...')
                    msg = self.unpacker.next()
                    pass # replaces next logging statement
                    #debug('received message: %s', msg)
                    msg_cb(msg)
                except StopIteration:
                    pass # replaces next logging statement
                    #debug('unpacker needs more data...')
                    break

        self.stream.loop_start(data_cb, error_cb)
        pass # replaces next logging statement
        #debug('exiting msgpack stream loop')


    def loop_stop(self):
        self.stream.loop_stop()
